from epsagon.trace import trace_factory
from moto import mock_events
import boto3
import json

fake_event = {
    "DetailType": 'test-detail-type',
    "Source": 'test-source',
    "Detail": json.dumps({"test": 1234}),
    "EventBusName": 'test-event-bus-name',
}

fake_event_bus_name_missing = {
    "DetailType": 'test-detail-type',
    "Source": 'test-source',
    "Detail": json.dumps({"test": 1234})
}


def setup_function(func):
    trace_factory.use_single_trace = True
    trace_factory.get_or_create_trace()


def teardown_function(func):
    trace_factory.singleton_trace = None

def _get_active_trace():
    return trace_factory.active_trace


def _put_event(client, event):
    response = client.put_events(Entries=[event])
    return response


@mock_events
def test_event_resources():
    client = boto3.client('events', region_name='us-west-1')
    _put_event(client, fake_event)
    trace = _get_active_trace()
    assert trace.events[0].resource["operation"] == 'PutEvents'
    assert trace.events[0].RESOURCE_TYPE == 'eventbridge'
    assert trace.events[0].resource["name"] == fake_event["EventBusName"]


@mock_events
def test_event_resources_without_bus_name():
    client = boto3.client('events', region_name='us-west-1')
    _put_event(client, fake_event_bus_name_missing)
    trace = _get_active_trace()
    assert trace.events[0].RESOURCE_TYPE == 'eventbridge'
    assert trace.events[0].resource["name"] == 'CloudWatch Events'


@mock_events
def test_event_metadata():
    client = boto3.client('events', region_name='us-west-1')
    _put_event(client, fake_event)
    trace = _get_active_trace()
    assert trace.events[0].resource["metadata"]['aws.cloudwatch.detail_type'] == fake_event["DetailType"]
    assert trace.events[0].resource["metadata"]['aws.cloudwatch.source'] == fake_event["Source"]
    assert trace.events[0].resource["metadata"]['aws.cloudwatch.detail'] == fake_event["Detail"]



import pytest
import urllib3
import epsagon.wrappers.python_function
from epsagon.trace import trace_factory
import epsagon.runners.python_function
import epsagon.constants


TEST_DOMAIN = 'jsonplaceholder.typicode.com'
TEST_PATH = '/todos/1'

def setup_function(func):
    trace_factory.use_single_trace = True
    epsagon.constants.COLD_START = True

def test_no_data_capture_with_urlopen(trace_transport):
    def use_urlopen():
        headers = {
            'Content-Type': 'application/json'
        }

        http = urllib3.PoolManager()
        conn = http.connection_from_url(TEST_DOMAIN)
        urllib_response = conn.urlopen(
            method='GET',
            url=TEST_PATH,
            body='',
            headers=headers,
            preload_content=False,
            decode_content=False,
        )
        return urllib_response

    response = use_urlopen()
    data = response.read()

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function():
        return use_urlopen()

    wrapped_response = wrapped_function()
    wrapped_data = wrapped_response.read()
    # We expect in this case that the data will be available in the buffer
    assert(len(wrapped_data) > 0)
    assert(len(wrapped_response.data) == 0)
    urllib3_event = trace_transport.last_trace.events[-1]
    # Payload should not be collected in the case of raw-stream usage.
    assert(urllib3_event.resource['metadata']['response_body'] is None)
    # Compare un-instrumented vs instrumented data
    assert (len(data) > 0)
    assert (len(response.data) == 0)
    assert(wrapped_data == data)
    assert(wrapped_response.data == response.data)

@pytest.mark.parametrize("preload_content", [True, False])
def test_data_capture_with_pool_manager(preload_content, trace_transport):
    def use_poolmanager():
        headers = {
            'Content-Type': 'application/json'
        }

        http = urllib3.PoolManager()
        return http.request(
            'GET', TEST_DOMAIN + TEST_PATH,
            headers=headers, body='',
            preload_content=preload_content
        )

    response = use_poolmanager()
    data = response.read()

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function():
        return use_poolmanager()

    wrapped_response = wrapped_function()
    wrapped_data = wrapped_response.read()
    if not preload_content:
        # In this case data will not be available in the buffer
        assert (len(wrapped_data) > 0)
        # But, data will not be stored in the `data` object
        assert (len(wrapped_response.data) == 0)
        # Compare un-instrumented vs instrumented data
        assert (len(data) > 0)
        assert (len(response.data) == 0)
    else:
        # In this case data will not be available in the buffer
        assert (len(wrapped_data) == 0)
        # But, data will be stored in the `data` object
        assert (len(wrapped_response.data) > 0)
        # Compare un-instrumented vs instrumented data
        assert (len(data) == 0)
        assert (len(response.data) > 0)
    assert(wrapped_data == data)
    assert(wrapped_response.data == response.data)



import json
import epsagon.wrappers.python_function
import epsagon.runners.python_function
import epsagon.constants
import mock
from kafka import KafkaProducer

TEST_URL = 'https://example.test/'

def record_mock(*args, **kwargs):
    return [{}, False, False]


@mock.patch('epsagon.trace.TraceFactory.add_event')
@mock.patch('kafka.producer.kafka.KafkaProducer._wait_on_metadata')
@mock.patch('kafka.producer.kafka.KafkaProducer._partition')
@mock.patch('kafka.producer.record_accumulator.RecordAccumulator.append', side_effect=record_mock)
def test_sanity(append_mock, partition_mock, wait_on_metadata_mock, add_event_mock):
    retval = 'success'
    body = {'test': 1}

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function():
        producer = KafkaProducer(
            bootstrap_servers=['host:10'],
            client_id='test_client_id',
            api_version=(0, 11, 5),
            value_serializer=lambda x: json.dumps(x).encode('ascii'),
        )
        response = producer.send('topic', body, headers=[('content-encoding', b'base64')])
        return retval
    assert wrapped_function() == retval
    wait_on_metadata_mock.assert_called()
    partition_mock.assert_called()
    append_mock.assert_called()
    add_event_mock.assert_called()
    event = add_event_mock.call_args_list[0].args[0]
    assert event.resource['name'] == 'topic'
    assert event.resource['operation'] == 'send'
    assert event.resource['type'] == 'kafka'
    assert event.resource['metadata']['messaging.kafka.client_id'] == (
        'test_client_id'
    )
    assert event.resource['metadata'][
       'messaging.message_payload_size_bytes'
   ] == len(str(body))
    assert event.resource['metadata']['messaging.message'] == body
    assert (
        epsagon.constants.EPSAGON_HEADER in
        event.resource['metadata']['messaging.headers']
    )


@mock.patch('epsagon.trace.TraceFactory.add_event')
@mock.patch('kafka.producer.kafka.KafkaProducer._wait_on_metadata')
@mock.patch('kafka.producer.kafka.KafkaProducer._partition')
@mock.patch('kafka.producer.record_accumulator.RecordAccumulator.append', side_effect=record_mock)
def test_no_header_injection(append_mock, partition_mock, wait_on_metadata_mock, add_event_mock):
    # Verify header is not injected in older kafka api versions (V1)
    retval = 'success'
    body = {'test': 1}

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function():
        producer = KafkaProducer(
            bootstrap_servers=['host:10'],
            client_id='test_client_id',
            api_version=(0, 10, 0),
            value_serializer=lambda x: json.dumps(x).encode('ascii'),
        )
        response = producer.send('topic', body)
        return retval
    assert wrapped_function() == retval
    wait_on_metadata_mock.assert_called()
    partition_mock.assert_called()
    append_mock.assert_called()
    add_event_mock.assert_called()
    event = add_event_mock.call_args_list[0].args[0]
    assert (
        epsagon.constants.EPSAGON_HEADER not in
        event.resource['metadata']
    )



"""
Main patcher module
"""

from __future__ import absolute_import
from importlib import import_module
import epsagon.modules


def _import_exists(module_name):
    """
    Validates if import module exists
    :param module_name: module name to import
    :return: Bool
    """
    try:
        import_module(module_name)
        return True
    except ImportError:
        return False


def patch_all():
    """
    Instrumenting all modules
    :return: None
    """
    for patch_module in epsagon.modules.MODULES:
        if _import_exists(patch_module):
            try:
                epsagon.modules.MODULES[patch_module].patch()
            except Exception:  # pylint: disable=broad-except
                pass



"""
A helper wrapper to measure internal functions duration
"""

from __future__ import absolute_import
import time
import functools
from ..trace import trace_factory


def measure(func):
    """A decorator to measure internal functions duration using labels."""

    @functools.wraps(func)
    def _measure(*args, **kwargs):
        """
        Creating a label based on the function name
        with the duration in seconds.
        """
        start_time = time.time()
        result = func(*args, **kwargs)
        trace_factory.add_label(
            '{}_duration'.format(func.__name__),
            float('{:.3f}'.format(time.time() - start_time))
        )
        return result
    return _measure



"""
Gunicorn patcher module.
"""

from __future__ import absolute_import
import warnings
import time
import uuid
import traceback

try:
    from gunicorn.config import (
        PreRequest, validate_callable, PostRequest, validate_post_request
    )
except ImportError:
    class PreRequest(object):
        """
        Dummy PreRequest Class
        """
        def __init__(self):
            pass

    class PostRequest(object):
        """
        Dummy Post Request Class
        """
        def __init__(self):
            pass

    def validate_callable(_):
        """
        Dummy Validate Callable Function
        :param _:
        :return:
        """
        return True

    def validate_post_request(_):
        """
        Dummy Validate Post Request Function
        :param _:
        :return:
        """
        return True

import epsagon
import epsagon.trace
from epsagon.common import EpsagonWarning
from epsagon.utils import collect_container_metadata


from ..event import BaseEvent
from ..http_filters import ignore_request
from ..utils import add_data_if_needed


class GunicornRunner(BaseEvent):
    """
    Represents Python Gunicorn event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'python_gunicorn'
    OPERATION = 'request'

    def __init__(self, start_time, request):
        """
        Initialize.
        :param start_time: event's start time (epoch).
        :param request: the incoming request.
        """
        super(GunicornRunner, self).__init__(start_time)

        self.event_id = str(uuid.uuid4())

        self.resource['name'] = 'localhost'
        if hasattr(request, 'headers'):
            headers = dict(request.headers)
            self.resource['name'] = (
                headers.get('HOST', 'localhost').split(':')[0]
            )
            add_data_if_needed(
                self.resource['metadata'],
                'Request Headers',
                headers
            )

        self.resource['operation'] = request.method
        self.resource['metadata'].update({'Path': request.path})
        self.resource['metadata'].update({'Query': request.query})

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: WSGI Response
        :return: None
        """
        add_data_if_needed(
            self.resource['metadata'],
            'Response Headers',
            dict(response.headers)
        )

        if response.status_code >= 500:
            self.set_error()


class PreRequestWrapped(PreRequest):
    """
    Represents Gunicorn Middleware's PreRequest Hook
    """
    name = 'pre_request'
    section = 'Server Hooks'
    validator = validate_callable(2)
    type = callable

    # pylint: disable=no-self-argument
    def pre_request(worker, req):
        """
        Runs before process of response.
        """
        if 'SyncWorker' not in type(worker).__name__:
            return

        trace = epsagon.trace.trace_factory.get_or_create_trace()
        trace.prepare()

        if ignore_request('', req.path.lower()):
            return

        # Create a Gunicorn runner with current request.
        try:
            runner = GunicornRunner(
                time.time(),
                req
            )
            trace.set_runner(runner)

            # Collect metadata in case this is a container.
            collect_container_metadata(runner.resource['metadata'])
        # pylint: disable=W0703
        except Exception as exception:
            # Regress to python runner.
            warnings.warn('Could not extract request', EpsagonWarning)
            epsagon.trace.trace_factory.add_exception(
                exception,
                traceback.format_exc()
            )

    # pylint: disable=no-staticmethod-decorator
    default = staticmethod(pre_request)
    desc = """\
        Called just before a worker processes the request.
        The callable needs to accept two instance variables for the Worker and
        the Request.
    """


class PostRequestWrapped(PostRequest):
    """
        Represents Gunicorn Middleware's PostRequest Hook
    """
    name = 'post_request'
    section = 'Server Hooks'
    validator = validate_post_request
    type = callable

    # pylint: disable=no-self-argument
    def post_request(worker, _req, environ, resp):
        """
        Runs after process of response.
        """
        if 'SyncWorker' not in type(worker).__name__:
            return

        # Ignoring non-relevant content types.
        if ignore_request(environ.get('Content-Type', '').lower(), ''):
            return

        trace = epsagon.trace.trace_factory.active_trace
        if trace and trace.runner:
            trace.runner.update_response(resp)
            epsagon.trace.trace_factory.send_traces()

    # pylint: disable=no-staticmethod-decorator
    default = staticmethod(post_request)
    desc = """\
        Called after a worker processes the request.
        The callable needs to accept two instance variables for the Worker and
        the Request.
    """



"""
Tornado AsyncHTTPClient events module.
"""

from __future__ import absolute_import

import functools
try:
    from urllib.parse import urlparse, urlunparse
except ImportError:
    from urlparse import urlparse, urlunparse
import traceback
from uuid import uuid4

from epsagon.utils import add_data_if_needed
from ..trace import trace_factory
from ..event import BaseEvent
from ..http_filters import (
    is_blacklisted_url,
    is_payload_collection_blacklisted
)
from ..utils import update_http_headers, normalize_http_url
from ..constants import HTTP_ERR_CODE, EPSAGON_HEADER_TITLE


class TornadoAsyncHTTPClientEvent(BaseEvent):
    """
    Represents base request event.
    """

    ORIGIN = 'tornado_client'
    RESOURCE_TYPE = 'http'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(TornadoAsyncHTTPClientEvent, self).__init__(start_time)
        self.event_id = 'tornado-client-{}'.format(str(uuid4()))

        request = args[0]
        headers = dict(request.headers)
        if headers:
            # Make sure trace ID is present in case headers will be removed.
            epsagon_trace_id = headers.get(EPSAGON_HEADER_TITLE)
            if epsagon_trace_id:
                self.resource['metadata']['http_trace_id'] = epsagon_trace_id

        parsed_url = urlparse(request.url)

        host_url = parsed_url.netloc.split(':')[0]
        full_url = urlunparse((
            parsed_url.scheme,
            host_url,
            parsed_url.path,
            parsed_url.params,
            parsed_url.query,
            parsed_url.fragment,
        ))

        self.resource['name'] = normalize_http_url(request.url)
        self.resource['operation'] = request.method
        self.resource['metadata']['url'] = request.url

        if not is_payload_collection_blacklisted(full_url):
            add_data_if_needed(
                self.resource['metadata'],
                'request_headers',
                headers
            )
            body = request.body
            if isinstance(body, bytes):
                body = body.decode('utf-8')
            if body:
                add_data_if_needed(
                    self.resource['metadata'],
                    'request_body',
                    body
                )

        if response is not None:
            callback = functools.partial(self.update_response)
            response.add_done_callback(callback)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def update_response(self, future):
        """
        Adds response data to event.
        :param future: Future response object
        :return: None
        """
        response = future.result()

        self.resource['metadata']['status_code'] = response.code
        self.resource = update_http_headers(
            self.resource,
            dict(response.headers)
        )

        full_url = self.resource['metadata']['url']

        if not is_payload_collection_blacklisted(full_url):
            add_data_if_needed(
                self.resource['metadata'],
                'response_headers',
                dict(response.headers)
            )
            body = response.body
            if isinstance(body, bytes):
                try:
                    body = body.decode('utf-8')
                except UnicodeDecodeError:
                    body = str(body)
            if body:
                add_data_if_needed(
                    self.resource['metadata'],
                    'response_body',
                    body
                )

        # Detect errors based on status code
        if response.code >= HTTP_ERR_CODE:
            self.set_error()


class TornadoClientEventFactory(object):
    """
    Factory class, generates AsyncHTTPClient event.
    """

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create an AsyncHTTPClient event.
        """
        # Detect if URL is blacklisted, and ignore.
        if is_blacklisted_url(args[0].url):
            return

        event = TornadoAsyncHTTPClientEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        trace_factory.add_event(event)



"""
Runner for a general python function
"""

from __future__ import absolute_import
import uuid
import json
import epsagon.trace
from ..event import BaseEvent
from ..trace_encoder import TraceEncoder


class PythonRunner(BaseEvent):
    """
    Represents general python event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'python_function'
    OPERATION = 'invoke'

    def __init__(
        self,
        start_time,
        wrapped_function,
        wrapped_args,
        wrapped_kwargs,
        name=None
    ):
        """
        Initialize.
        :param start_time: event's start time (epoch).
        :param wrapped_function: the function this runner is wrapping.
        :param wrapped_args: the arguments the function was called with.
        :param wrapped_kwargs: the keyword arguments the function was
               called with.
        """

        super(PythonRunner, self).__init__(start_time)

        self.event_id = str(uuid.uuid4())
        self.resource['name'] = name if name else wrapped_function.__name__
        self.resource['operation'] = self.OPERATION

        self.resource['metadata'].update({
            'python.module': wrapped_function.__module__,
            'python.function.name': wrapped_function.__name__,
        })

        if wrapped_args:
            self.add_json_field('python.function.args', wrapped_args)
            self.resource['metadata']['python.function.args_length'] = len(
                wrapped_args
            )

        if wrapped_kwargs:
            self.add_json_field('python.function.kwargs', wrapped_kwargs)
            self.resource['metadata']['python.function.kwargs_length'] = len(
                wrapped_kwargs
            )

    def add_json_field(self, name, data):
        """
        Add a field to metadata with value `data` and name `name`,
            only if it is JSON serializable
        """
        trace = epsagon.trace.trace_factory.get_trace()
        if not trace or trace.metadata_only:
            return

        try:
            json.dumps(data, cls=TraceEncoder, ensure_ascii=True)
            self.resource['metadata'][name] = data
        except TypeError:
            pass



"""
sqlalchemy patcher module
"""
from __future__ import absolute_import

from epsagon.modules.general_wrapper import wrapper
from ..events.sqlalchemy import SqlAlchemyEventFactory
from ..utils import patch_once


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for sqlalchemy instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    return wrapper(SqlAlchemyEventFactory, wrapped, instance, args, kwargs)

def patch():
    """
    patch module.
    :return: None
    """

    patch_once(
        'sqlalchemy.orm.session',
        'Session.__init__',
        _wrapper
    )

    patch_once(
        'sqlalchemy.orm.session',
        'Session.close',
        _wrapper
    )



"""
kafka-python patcher module.
"""

from __future__ import absolute_import
import wrapt
from epsagon.modules.general_wrapper import wrapper
from ..events.kafka import KafkaEventFactory
from ..constants import EPSAGON_HEADER
from ..utils import get_epsagon_http_trace_id


def _parse_args(
    topic,
    value=None,
    key=None,
    headers=None,
    partition=None,
    timestamp_ms=None
):
    """Sort and return args and kwargs according to the original signature"""
    return (topic, ), {
        'value': value,
        'key': key,
        'headers': headers,
        'partition': partition,
        'timestamp_ms': timestamp_ms
    }


def _wrapper(wrapped, instance, args, kwargs):
    """KafkaProducer.send wrapper"""
    new_args, new_kwargs = _parse_args(*args, **kwargs)

    # Adds epsagon header only on Kafka record V2. V0/V1 don't support it
    # pylint: disable=protected-access
    if instance._max_usable_produce_magic() == 2:
        if not new_kwargs.get('headers'):
            new_kwargs['headers'] = []
        new_kwargs['headers'].append(
            (EPSAGON_HEADER, get_epsagon_http_trace_id().encode())
        )

    return wrapper(KafkaEventFactory, wrapped, instance, new_args, new_kwargs)


def patch():
    """
    patch module.
    :return: None
    """
    wrapt.wrap_function_wrapper(
        'kafka.producer.kafka',
        'KafkaProducer.send',
        _wrapper
    )



"""
Epsagon generic wrapper used only in Lambda environments.
"""

from .utils import init as epsagon_init, import_original_module
from .wrappers import lambda_wrapper


def init_module():
    """
    Initialize user's module handler.
    :return: wrapper handler.
    """
    original_module, module_path, handler_name = import_original_module()
    try:
        handler = original_module
        for name in module_path.split('.')[1:] + [handler_name]:
            handler = getattr(handler, name)
        return handler
    except AttributeError:
        raise AttributeError(
            'No handler {} in module {}'.format(handler_name, module_path)
        )


epsagon_init()
wrapper = lambda_wrapper(init_module())



""" JSONEncoder for trace objects """

from datetime import datetime, date
import json


class TraceEncoder(json.JSONEncoder):
    """
    An encoder for the trace json
    """

    def default(self, o):  # pylint: disable=method-hidden
        if isinstance(o, set):
            return list(o)
        if isinstance(o, (datetime, date)):
            return o.isoformat()
        if isinstance(o, bytes):
            return o.decode('utf-8', errors='ignore')

        output = repr(o)
        try:
            output = json.JSONEncoder.default(self, o)
        except TypeError:
            pass
        return output



"""
PynamoDB events module.
"""

from __future__ import absolute_import
from uuid import uuid4
import json
from ..trace import trace_factory
from .botocore import BotocoreDynamoDBEvent


class NestedObject(object):
    """
    Creating a nested object based on a dict.
    """
    def __init__(self, **data):
        for k, v in data.items():
            if isinstance(v, dict):
                self.__dict__[k] = NestedObject(**v)
            else:
                self.__dict__[k] = v


class PynamoDBVendoredEventAdapter(object):
    """
    Factory class, generates PynamoDB event from botocore vendored.
    """

    @staticmethod
    def create_event(wrapped, _instance, args, kwargs, start_time, response,
                     exception):
        """Creates DynamoDB event based on PynamoDB data"""
        new_args = (
            args[0].headers['X-Amz-Target'].decode('utf-8').split('.')[1],
            json.loads(args[0].body.decode('utf-8'))
        )

        new_instance = NestedObject(**{
            'meta': {
                'region_name': args[0].url.split('.')[1]
            }
        })

        new_response = {
            'ResponseMetadata': {
                'RequestId': response.headers['x-amzn-requestid'],
                'HTTPStatusCode': response.status_code,
                'RetryAttempts': None,
            },
        }
        new_response.update(response.json())
        event = BotocoreDynamoDBEvent(
            wrapped,
            new_instance,
            new_args,
            kwargs,
            start_time,
            new_response,
            exception
        )
        event.origin = 'pynamodb'
        event.resource['metadata'].pop('Retry Attempts')

        trace_factory.add_event(event)


class PynamoDBEventAdapter(object):
    """
    Factory class, generates PynamoDB event from pynamodb module.
    """

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """Creates DynamoDB event based on PynamoDB data"""
        new_response = {
            'ResponseMetadata': {
                'RequestId': 'pynamodb-{}'.format(str(uuid4())),
                'HTTPStatusCode': 200 if exception is None else 500,
                'RetryAttempts': None,
            },
        }
        if response:
            new_response.update(response)
        event = BotocoreDynamoDBEvent(
            wrapped,
            instance.client,
            args,
            kwargs,
            start_time,
            new_response,
            exception
        )
        event.origin = 'pynamodb'
        event.resource['metadata'].pop('Retry Attempts')

        trace_factory.add_event(event)



"""
Runner for Google Functions.

"""

from __future__ import absolute_import
import os
from uuid import uuid4
from ..event import BaseEvent
from .. import constants


class GoogleFunctionRunner(BaseEvent):
    """
    Represents Google function event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'google_function'
    OPERATION = 'Invoke'

    def __init__(self, start_time):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        """

        super(GoogleFunctionRunner, self).__init__(start_time)

        self.event_id = 'gcp_{}'.format(str(uuid4()))
        self.resource['name'] = os.getenv(
            'FUNCTION_NAME',
            ''
        )
        self.resource['operation'] = self.OPERATION

        self.resource['metadata'].update({
            'gcp_project': os.getenv('GCP_PROJECT', ''),
            'function_version': os.getenv('X_GOOGLE_FUNCTION_VERSION', ''),
            'memory':  os.getenv('FUNCTION_MEMORY_MB', ''),
            'cold_start': constants.COLD_START,
            'region': os.getenv('FUNCTION_REGION', ''),
            'supervisor_hostname': os.getenv('SUPERVISOR_HOSTNAME', ''),
            'supervisor_internal_port': os.getenv(
                'SUPERVISOR_INTERNAL_PORT', ''
            ),
            'virtual_env': os.getenv('VIRTUAL_ENV', ''),
            'worker_port': os.getenv('WORKER_PORT', ''),
        })



"""
HTTP triggers for frameworks.
"""

from __future__ import absolute_import
from ..event import BaseEvent


class BaseHTTPTrigger(BaseEvent):
    """
    Represents base HTTP trigger
    """
    ORIGIN = 'trigger'


class SQSHTTPTrigger(BaseHTTPTrigger):
    """
    Represents SQS HTTP trigger
    """
    RESOURCE_TYPE = 'sqs'

    # pylint: disable=W0613
    def __init__(self, start_time, request):
        """
        Initialize.
        :param start_time: event's start time (epoch).
        :param request: HTTP request.
        """

        super(SQSHTTPTrigger, self).__init__(start_time)

        self.event_id = request.headers.get('X-Aws-Sqsd-Msgid')
        self.resource['name'] = request.headers.get('X-Aws-Sqsd-Queue', 'N/A')
        self.resource['operation'] = 'ReceiveMessage'


class HTTPTriggerFactory(object):
    """
    Represents a HTTP Trigger Factory.
    """
    FACTORY = {
        class_obj.RESOURCE_TYPE: class_obj
        for class_obj in BaseHTTPTrigger.__subclasses__()
    }

    @staticmethod
    def factory(start_time, request):
        """
        Creates trigger event object.
        :param start_time: event's start time (epoch)
        :param request: HTTP request.
        :return: Event or None.
        """

        if request.headers.get('X-Aws-Sqsd-Msgid'):
            return SQSHTTPTrigger(start_time, request)
        return None



import os
import threading
import pytest
import mock
from flask import Flask, request
from epsagon import trace_factory
from epsagon.wrappers.flask import FlaskWrapper
from time import sleep
from .common import multiple_threads_handler

# Setting demo Flask app
RETURN_VALUE = 'a'
app_test = Flask('test')
FlaskWrapper(app_test)


def setup_function(func):
    trace_factory.use_single_trace = False


@app_test.route('/')
def test():
    return RETURN_VALUE


@app_test.route('/a')
def a_route():
    sleep(0.1)
    return "a"


@app_test.route('/b')
def b_route():
    return "b"


@app_test.route('/error')
def error():
    raise Exception('test')


@app_test.route('/multiple_threads')
def multiple_threads_route():
    multiple_threads_handler()
    return "multiple_threads"


@pytest.fixture
def client():
    """ A test client that has the Authorization header """
    app_test.testing = True
    app = app_test.test_client()
    return app


CLIENT = app_test.test_client()


@app_test.route('/self_call', methods=["POST"])
def self_call():
    json_data = request.get_json()
    inner_call = None
    if json_data:
        inner_call = json_data['inner_call']

    if not inner_call:
        return CLIENT.post('/self_call', json={
            'inner_call': 'True'
        })

    return RETURN_VALUE

###


@mock.patch('warnings.warn')
@mock.patch('epsagon.triggers.http.HTTPTriggerFactory')
@mock.patch('epsagon.runners.flask.FlaskRunner')
@mock.patch('time.time', return_value=1)
def test_flask_wrapper_before_request(_, runner_mock, trigger_mock, __, client):
    """Test runner and trigger init on new request."""
    result = client.get('/')

    runner_mock.assert_called_with(1, app_test, request)
    trigger_mock.factory.assert_called_once()

    assert result.data.decode('ascii') == RETURN_VALUE


@mock.patch('warnings.warn')
@mock.patch('epsagon.runners.flask.FlaskRunner.update_response')
def test_flask_wrapper_after_request(runner_mock, _, client):
    """Test update response called on new request."""
    client.get('/')

    runner_mock.assert_called_once()


@mock.patch('warnings.warn')
@mock.patch('epsagon.trace.trace_factory.get_or_create_trace')
@mock.patch('epsagon.trace.trace_factory.get_trace')
def test_flask_wrapper_teardown_request(get_trace_mock, create_trace_mock, _, client):
    """Test tracer gets new event and send it on new request."""
    client.get('/')
    create_trace_mock().set_runner.assert_called_once()
    get_trace_mock().send_traces.assert_called_once()


@mock.patch('warnings.warn')
@mock.patch('epsagon.runners.flask.FlaskRunner.set_exception')
def test_flask_wrapper_teardown_exception(exception_mock, _, client):
    """Test runner gets an exception on error request."""
    with pytest.raises(Exception):
        client.get('/error')

        exception_mock.assert_called_once()


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
def test_lambda_wrapper_multi_thread(_):
    assert not trace_factory.use_single_trace


def validate_response(role, result, trace_transport):
    """
    Validate Epsagon trace was generated the right way,
    and that the response matches what we expect
    """
    assert trace_transport.last_trace.events[0].resource[
        'metadata']['Response Data'].decode('ascii') == role
    assert result.data.decode('ascii') == role


def thread_client(role, trace_transport, client):
    """
    Get a string from an endpoint, and validate Epsagon
    trace and response
    """
    result = client.get('/{0}'.format(role))
    validate_response(role, result, trace_transport)


def test_flask_wrapper_multiple_requests(trace_transport, client):
    """
    Make 2 simulatanous requests.
    Make sure none of the responses or generated traces mix up.
    """
    a = threading.Thread(target=thread_client, args=(
        'a', trace_transport, client))
    b = threading.Thread(target=thread_client, args=(
        'b', trace_transport, client))

    for thread in [a, b]:
        thread.start()

    for thread in [a, b]:
        thread.join()


def test_ignore_response_data(trace_transport, client):
    """
    Make sure no response data is being captured when
    Make sure none of the responses or generated traces mix up.
    """
    os.environ['EPSAGON_IGNORE_FLASK_RESPONSE'] = 'TRUE'
    client.get('/a')
    assert (
        'Response Data' not in
        trace_transport.last_trace.events[0].resource['metadata']
    )
    os.environ.pop('EPSAGON_IGNORE_FLASK_RESPONSE')


def test_call_to_self(trace_transport, client):
    """
    And API that calls itself. Make sure instrumentation doesn't throw
    and exception that gets to the user.
    """
    result = CLIENT.post('/self_call')
    assert result.data.decode('ascii') == RETURN_VALUE


def test_flask_wrapper_route_multiple_threads(trace_transport, client):
    """
    Makes a request to a route which invokes multiple threads
    perfoming http requests.
    Make sure no trace is created for those threads.
    """
    role = 'multiple_threads'
    result = client.get('/{}'.format(role))
    validate_response(role, result, trace_transport)
    # validating no `zombie` traces exist
    assert not trace_factory.traces



"""Common objects"""


class ErrorCode(object):
    """
    Error codes enum
    """
    OK = 0
    ERROR = 1
    EXCEPTION = 2
    TIMEOUT = 3


class EpsagonWarning(Warning):
    """
    An Epsagon warning.
    """



"""
Tracing route for Python fastapi.
"""
import time
import uuid
import json
import json.decoder
import asyncio

import warnings
from fastapi import Request, Response
from starlette.requests import ClientDisconnect
from starlette.concurrency import run_in_threadpool

import epsagon.trace
from epsagon.runners.fastapi import FastapiRunner
from epsagon.common import EpsagonWarning
from epsagon.constants import EPSAGON_MARKER
from epsagon.utils import (
    collect_container_metadata,
    get_traceback_data_from_exception
)
from ..http_filters import ignore_request
from ..utils import is_lambda_env, print_debug

DEFAULT_SUCCESS_STATUS_CODE = 200
DEFAULT_ERROR_STATUS_CODE = 500
EPSAGON_REQUEST_PARAM_NAME = 'epsagon_request'
SCOPE_UNIQUE_ID = 'trace_unique_id'
SCOPE_CONTAINER_METADATA_COLLECTED = 'container_metadata'
SCOPE_IGNORE_REQUEST = 'ignore_request'

def _handle_wrapper_params(_args, kwargs, original_request_param_name):
    """
    Handles the sync/async given parameters - gets the request object
    If original handler is set to get the Request object, then getting the
    request using this param. Otherwise, trying to get the Request object using
    Epsagon injected param (and removing this injected param)
    :return: the request object, None if not exists
    """
    if original_request_param_name and original_request_param_name in kwargs:
        return kwargs[original_request_param_name]
    return kwargs.pop(EPSAGON_REQUEST_PARAM_NAME, None)


def _handle_response(epsagon_scope, response, status_code, trace, raised_err):
    """
    Handles the HTTP handler response.
    :param epsagon_scope:
    :param response:
    :param status_code:
    :param trace: to update the response with
    :param raised_err: any error which might occured while trying to get
    the response.
    :return: the response, raising raised_err if its not None.
    """
    try:
        if raised_err:
            status_code = DEFAULT_ERROR_STATUS_CODE
        elif isinstance(response, Response):
            if ignore_request(
                    response.headers.get('Content-Type', '').lower(),
                    ''
            ):
                epsagon_scope[SCOPE_IGNORE_REQUEST] = True
                return response

        if response is not None:
            trace.runner.update_response(response, status_code=status_code)

    except Exception as exception:  # pylint: disable=W0703
        print_debug('Failed to handle response: {}'.format(exception))

    if raised_err:
        raise raised_err

    return response


def _get_epsagon_scope_data(request: Request) -> str:
    """
    Gets the epsagon scope data from a given request.
    Returns None if not found.
    """
    try:
        if not request.scope:
            return None
        return request.scope.get(EPSAGON_MARKER)
    except Exception: # pylint: disable=broad-except
        return None


def _extract_request_body(trace, request):
    """
    Extracts the request body (if exists), and saves it to the trace runner.
    """
    try:
        loop = asyncio.new_event_loop()
        trace.runner.update_request_body(
            json.dumps(loop.run_until_complete(request.json()))
        )
    except json.decoder.JSONDecodeError:
        pass
    except ClientDisconnect:
        print_debug(
            'Could not extract request body - client is disconnected'
        )
    except Exception as exception: # pylint: disable=W0703
        print_debug(
            'Could not extract request body: {}'.format(exception)
        )
    finally:
        if loop:
            loop.close()


def _setup_handler(request):
    """
    Setup the handler according to given request epsa.
    Extracts and returns the epsagon scope & epsagon trace.
    :return: A tuple of (epsagon_scope, trace)
    """
    epsagon_scope = _get_epsagon_scope_data(request)
    if not epsagon_scope:
        return None, None
    unique_id = epsagon_scope.get(SCOPE_UNIQUE_ID)
    if not unique_id or epsagon_scope.get(SCOPE_IGNORE_REQUEST):
        return None, None

    epsagon.trace.trace_factory.set_thread_local_unique_id(
        unique_id=unique_id
    )
    return epsagon_scope, epsagon.trace.trace_factory.get_trace()


def _setup_trace_runner(epsagon_scope, trace, request):
    """
    Setups the trace runner event - creates the FastAPI runner event,
    collects container metadata (if relevant).
    Assumption: a runner event has NOT been created before for given trace.
    :return: True if succeeded to setup runner event, False otherwise
    """
    try:
        trace.set_runner(FastapiRunner(time.time(), request))
    except Exception: # pylint: disable=W0703
        print_debug('Failed to add FastAPI runner event, skipping trace')
        # failed to add runner event, skipping trace
        return False
    try:
        if not epsagon_scope.get(SCOPE_CONTAINER_METADATA_COLLECTED):
            collect_container_metadata(trace.runner.resource['metadata'])
            epsagon_scope[SCOPE_CONTAINER_METADATA_COLLECTED] = True
    except Exception: # pylint: disable=W0703
        warnings.warn(
            'Could not extract container metadata',
            EpsagonWarning
        )
    return True


def _fastapi_handler(
        original_handler,
        request,
        status_code,
        args,
        kwargs
):
    """
    FastAPI generic handler - for callbacks executed by a threadpool
    :param original_handler: the wrapped original handler
    :param request: the given handler request
    :param status_code: the default configured response status code.
    Can be None when called by exception handlers wrapper, as there's
    no status code configuration for exception handlers.
    """
    has_setup_succeeded = False
    should_ignore_request = False
    try:
        epsagon_scope, trace = _setup_handler(request)
        if epsagon_scope and trace:
            has_setup_succeeded = True
        if ignore_request('', request.url.path.lower()):
            should_ignore_request = True
            epsagon_scope[SCOPE_IGNORE_REQUEST] = True

    except Exception: # pylint: disable=broad-except
        has_setup_succeeded = False

    if not has_setup_succeeded or should_ignore_request:
        return original_handler(*args, **kwargs)

    created_runner = False
    response = None
    if not trace.runner:
        if not _setup_trace_runner(epsagon_scope, trace, request):
            return original_handler(*args, **kwargs)

    raised_err = None
    try:
        response = original_handler(*args, **kwargs)
    except Exception as exception:  # pylint: disable=W0703
        raised_err = exception
    finally:
        try:
            epsagon.trace.trace_factory.unset_thread_local_unique_id()
        except Exception: # pylint: disable=broad-except
            pass
    # no need to update request body if runner already created before
    if created_runner:
        _extract_request_body(trace, request)

    return _handle_response(
        epsagon_scope,
        response,
        status_code,
        trace,
        raised_err
    )


# pylint: disable=too-many-statements
def _wrap_handler(dependant, status_code):
    """
    Wraps the endppint handler.
    """
    original_handler = dependant.call
    is_async = asyncio.iscoroutinefunction(original_handler)
    if is_async:
        # async endpoints are not supported
        return

    original_request_param_name = dependant.request_param_name
    if not original_request_param_name:
        dependant.request_param_name = EPSAGON_REQUEST_PARAM_NAME

    def wrapped_handler(*args, **kwargs):
        """
        Synchronous wrapper handler
        """
        request: Request = _handle_wrapper_params(
            args, kwargs, original_request_param_name
        )
        return _fastapi_handler(
            original_handler, request, status_code, args, kwargs
        )

    dependant.call = wrapped_handler


def route_class_wrapper(wrapped, instance, args, kwargs):
    """
    Route class wrapper - traces each route request & response.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    """
    result = wrapped(*args, **kwargs)
    # Skip on Lambda environment since it's not relevant and might be duplicate
    if not is_lambda_env() and instance:
        try:
            if instance.dependant and instance.dependant.call:
                _wrap_handler(
                    instance.dependant,
                    kwargs.get('status_code', DEFAULT_SUCCESS_STATUS_CODE)
                )
        except Exception: # pylint: disable=broad-except
            pass

    return result


def exception_handler_wrapper(original_handler):
    """
    Wraps an exception handler
    """
    is_async = asyncio.iscoroutinefunction(original_handler)
    if is_async:
        # async handlers are not supported
        return original_handler

    def wrapped_handler(*args, **kwargs):
        if not args or len(args) != 2:
            return original_handler(*args, **kwargs)
        request: Request = args[0]
        return _fastapi_handler(
            original_handler, request, None, args, kwargs
        )

    return wrapped_handler


def _clean_trace(trace):
    """ Cleans the given trace """
    if trace:
        try:
            epsagon.trace.trace_factory.pop_trace(trace=trace)
        except Exception: # pylint: disable=broad-except
            pass


async def server_call_wrapper(wrapped, _instance, args, kwargs):
    """
    Wraps the main fastapi server request entrypoint - which is
    the ServerErrorMiddleware __call_ function.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    """
    # Skip on Lambda environment since it's not relevant and might be duplicate
    if is_lambda_env() or not args or len(args) != 3:
        return await wrapped(*args, **kwargs)

    scope = args[0]
    if not scope or scope.get('type', '') != 'http':
        return await wrapped(*args, **kwargs)

    trace = None
    try:
        epsagon.trace.trace_factory.switch_to_multiple_traces()
        unique_id = str(uuid.uuid4())
        trace = epsagon.trace.trace_factory.get_or_create_trace(
            unique_id=unique_id
        )
        trace.prepare()
        scope[EPSAGON_MARKER] = {
            SCOPE_UNIQUE_ID: unique_id,
        }
    except Exception: # pylint: disable=broad-except
        _clean_trace(trace)
        return await wrapped(*args, **kwargs)

    response = None
    raised_error = None
    sent_trace = False
    epsagon_scope = scope[EPSAGON_MARKER]
    try:
        response = await wrapped(*args, **kwargs)
    except Exception as exception: # pylint: disable=broad-except
        raised_error = exception

    if trace.runner and not epsagon_scope.get(SCOPE_IGNORE_REQUEST):
        try:
            if raised_error:
                traceback_data = get_traceback_data_from_exception(raised_error)
                trace.runner.set_exception(raised_error, traceback_data)
                trace.runner.update_status_code(
                    DEFAULT_ERROR_STATUS_CODE,
                    override=False
                )
            await run_in_threadpool(
                epsagon.trace.trace_factory.send_traces,
                trace=trace
            )
            sent_trace = True
        except Exception as exception: # pylint: disable=broad-except
            print_debug('Failed to send traces: {}'.format(exception))

    scope.pop(EPSAGON_MARKER, None)
    if not sent_trace:
        _clean_trace(trace)
    if raised_error:
        raise raised_error from None
    return response



"""
FastAPI wrapper tests
"""
import time
import pytest
import asynctest
import asyncio
from typing import List
from httpx import AsyncClient
from pydantic import BaseModel
from fastapi import FastAPI, APIRouter, Request
from fastapi.routing import APIRoute
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder
from epsagon import trace_factory
from epsagon.common import ErrorCode
from epsagon.runners.fastapi import FastapiRunner
from epsagon.wrappers.fastapi import (
    DEFAULT_SUCCESS_STATUS_CODE,
    DEFAULT_ERROR_STATUS_CODE,
)
from .common import multiple_threads_handler

RETURN_VALUE = 'testresponsedata'
ROUTER_RETURN_VALUE = 'router-endpoint-return-data'
CUSTOM_ROUTE_RETURN_VALUE = 'custom-route-endpoint-return-data'
REQUEST_OBJ_PATH = '/given_request'
TEST_ROUTER_PREFIX = '/test-router-path'
TEST_ROUTER_PATH = '/test-router'
TEST_CUSTOM_ROUTE_PATH = '/atest-custom-route'
TEST_CUSTOM_ROUTE_PREFIX = '/test-custom-route-main'
MULTIPLE_THREADS_KEY = "multiple_threads"
MULTIPLE_THREADS_ROUTE = f'/{MULTIPLE_THREADS_KEY}'
MULTIPLE_THREADS_RETURN_VALUE = MULTIPLE_THREADS_KEY
TEST_POST_DATA = {'post_test': '123'}
CUSTOM_RESPONSE = ["A"]
CUSTOM_RESPONSE_PATH = "/custom_response"
BASE_MODEL_RESPONSE_PATH = "/base_model_response"
CUSTOM_STATUS_CODE = 202
CUSTOM_STATUS_CODE_PATH = "/custom_status_code"
OVERRIDDEN_CUSTOM_STATUS_CODE_PATH = "/overridden_custom_status_code"
CUSTON_EXCEPTION_HANDLER_RESPONSE = {"some exception": "test"}
DEFAULT_EXCEPTION_HANDLER_RESPONSE = {"default exception": "test2"}
HANDLED_EXCEPTION_PATH = "/custom_handled_exception"
UNHANDLED_EXCEPTION_PATH = "/default_handled_exception"


class CustomBaseModel(BaseModel):
    data: List[str]

class CustomRouteClass(APIRoute):
    pass

def _get_response_data(key):
    return {key: key}

def _get_response(key):
    return JSONResponse(content=_get_response_data(key))

# test fastapi app handlers
def handle():
    return _get_response(RETURN_VALUE)

def handle_custom_response(response_model=List[str]):
    return CUSTOM_RESPONSE

def handle_base_model_response(response_model=CustomBaseModel):
    return CustomBaseModel(data=CUSTOM_RESPONSE)

def handle_custom_status_code(response_model=List[str]):
    return CUSTOM_RESPONSE

def handle_overridden_custom_status_code():
    return _get_response(RETURN_VALUE)

def handle_given_request(request: Request):
    assert request.method == 'POST'
    loop = None
    try:
        loop = asyncio.new_event_loop()
        assert loop.run_until_complete(request.json()) == TEST_POST_DATA
    finally:
        if loop:
            loop.close()
    return _get_response(RETURN_VALUE)

def handle_a():
    time.sleep(0.2)
    return _get_response('a')

def handle_b():
    return _get_response('b')

def handle_router_endpoint():
    return _get_response(ROUTER_RETURN_VALUE)

def handle_custom_route_endpoint():
    return _get_response(CUSTOM_ROUTE_RETURN_VALUE)

def multiple_threads_route():
    multiple_threads_handler()
    return _get_response(MULTIPLE_THREADS_RETURN_VALUE)


class CustomFastAPIException(Exception):
    pass


class HandledFastAPIException(Exception):
    pass


class UnhandledFastAPIException(Exception):
    pass


def handle_error_from_route():
    raise CustomFastAPIException('test')


def handle_raise_custom_error():
    raise HandledFastAPIException('test')


def handle_raise_unhandled_error():
    raise UnhandledFastAPIException('test')


def custom_exception_handler(request: Request, exc):
    return JSONResponse(
        status_code=CUSTOM_STATUS_CODE,
        content=CUSTON_EXCEPTION_HANDLER_RESPONSE
    )


def default_exception_handler(request: Request, exc):
    return JSONResponse(
        content=DEFAULT_EXCEPTION_HANDLER_RESPONSE
    )


def _build_fastapi_app():
    app = FastAPI()
    app.add_api_route("/", handle, methods=["GET"])
    app.add_api_route(
        CUSTOM_RESPONSE_PATH,
        handle_custom_response,
        methods=["GET"]
    )
    app.add_api_route(
        BASE_MODEL_RESPONSE_PATH,
        handle_base_model_response,
        methods=["GET"]
    )
    app.add_api_route(
        CUSTOM_STATUS_CODE_PATH,
        handle_custom_status_code,
        methods=["GET"],
        status_code=CUSTOM_STATUS_CODE
    )
    app.add_api_route(
        OVERRIDDEN_CUSTOM_STATUS_CODE_PATH,
        handle_overridden_custom_status_code,
        methods=["GET"],
        status_code=CUSTOM_STATUS_CODE
    )
    app.add_api_route(REQUEST_OBJ_PATH, handle_given_request, methods=["POST"])
    app.add_api_route("/a", handle_a, methods=["GET"])
    app.add_api_route("/b", handle_b, methods=["GET"])
    app.add_api_route("/err", handle_error_from_route, methods=["GET"], status_code=200)
    app.add_api_route(MULTIPLE_THREADS_ROUTE, multiple_threads_route, methods=["GET"])
    router = APIRouter()
    router.add_api_route(TEST_ROUTER_PATH, handle_router_endpoint)
    app.include_router(router, prefix=TEST_ROUTER_PREFIX)
    router_with_custom_route = APIRouter(route_class=CustomRouteClass)
    router_with_custom_route.add_api_route(TEST_CUSTOM_ROUTE_PATH, handle_custom_route_endpoint)
    app.include_router(router_with_custom_route, prefix=TEST_CUSTOM_ROUTE_PREFIX)
    return app

@pytest.fixture(scope='function', autouse=False)
def fastapi_app():
    return _build_fastapi_app()


@pytest.fixture(scope='function', autouse=False)
def fastapi_app_with_exception_handlers(fastapi_app):
    app = fastapi_app
    app.add_api_route(
        HANDLED_EXCEPTION_PATH, handle_raise_custom_error, methods=["GET"]
    )
    app.add_api_route(
        UNHANDLED_EXCEPTION_PATH, handle_raise_unhandled_error, methods=["GET"]
    )
    app.add_exception_handler(
        HandledFastAPIException,
        custom_exception_handler,
    )
    app.add_exception_handler(
        Exception,
        default_exception_handler,
    )
    return app


@pytest.mark.asyncio
async def test_fastapi_sanity(trace_transport, fastapi_app):
    """Sanity test."""
    async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
        response = await ac.get("/?x=testval")
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == '/'
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    expected_response_data = _get_response_data(RETURN_VALUE)
    assert runner.resource['metadata']['Response Data'] == (
        expected_response_data
    )
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert response_data == expected_response_data
    # validating no `zombie` traces exist
    assert not trace_factory.traces


@pytest.mark.asyncio
async def test_fastapi_custom_response(trace_transport, fastapi_app):
    """custom response test."""
    request_path = f'{CUSTOM_RESPONSE_PATH}?x=testval'
    async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
        response = await ac.get(request_path)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == CUSTOM_RESPONSE_PATH
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert runner.resource['metadata']['Response Data'] == (
        jsonable_encoder(CUSTOM_RESPONSE)
    )
    assert response_data == CUSTOM_RESPONSE
    # validating no `zombie` traces exist
    assert not trace_factory.traces


@pytest.mark.asyncio
async def test_fastapi_base_model_response(trace_transport, fastapi_app):
    """base model response test."""
    request_path = f'{BASE_MODEL_RESPONSE_PATH}?x=testval'
    async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
        response = await ac.get(request_path)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    expected_response_data = CustomBaseModel(data=CUSTOM_RESPONSE)
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == BASE_MODEL_RESPONSE_PATH
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert runner.resource['metadata']['Response Data'] == (
        jsonable_encoder(expected_response_data)
    )
    assert response_data == expected_response_data
    # validating no `zombie` traces exist
    assert not trace_factory.traces


@pytest.mark.asyncio
async def test_fastapi_custom_status_code(trace_transport, fastapi_app):
    """custom status code test."""
    request_path = f'{CUSTOM_STATUS_CODE_PATH}?x=testval'
    async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
        response = await ac.get(request_path)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == CUSTOM_STATUS_CODE_PATH
    assert runner.resource['metadata']['status_code'] == CUSTOM_STATUS_CODE
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert runner.resource['metadata']['Response Data'] == (
        jsonable_encoder(CUSTOM_RESPONSE)
    )
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert runner.resource['metadata']['Response Data'] == (
        jsonable_encoder(CUSTOM_RESPONSE)
    )
    assert response_data == CUSTOM_RESPONSE
    # validating no `zombie` traces exist
    assert not trace_factory.traces


@pytest.mark.asyncio
async def test_fastapi_custom_status_code_overridden(trace_transport, fastapi_app):
    """custom status code test - status code overridden by returned Response """
    path = OVERRIDDEN_CUSTOM_STATUS_CODE_PATH
    request_path = f'{path}?x=testval'
    async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
        response = await ac.get(request_path)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == path
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    expected_response_data = _get_response_data(RETURN_VALUE)
    assert runner.resource['metadata']['Response Data'] == (
        expected_response_data
    )
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert response_data == expected_response_data
    # validating no `zombie` traces exist
    assert not trace_factory.traces


@pytest.mark.asyncio
async def test_fastapi_given_request(trace_transport, fastapi_app):
    """handler with a request parameter test."""
    request_path = f'{REQUEST_OBJ_PATH}?x=testval'
    async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
        response = await ac.post(request_path, json=TEST_POST_DATA)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == REQUEST_OBJ_PATH
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    expected_response_data = _get_response_data(RETURN_VALUE)
    assert runner.resource['metadata']['Response Data'] == (
        expected_response_data
    )
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert response_data == expected_response_data
    # validating no `zombie` traces exist
    assert not trace_factory.traces


@pytest.mark.asyncio
async def test_fastapi_custom_router(trace_transport, fastapi_app):
    """Custom router sanity test."""
    full_route_path= f'{TEST_ROUTER_PREFIX}{TEST_ROUTER_PATH}'
    async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
        response = await ac.get(full_route_path)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == full_route_path
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    expected_response_data = _get_response_data(ROUTER_RETURN_VALUE)
    assert runner.resource['metadata']['Response Data'] == (
        expected_response_data
    )
    assert response_data == expected_response_data


@pytest.mark.asyncio
async def test_fastapi_custom_api_route(trace_transport, fastapi_app):
    """Custom api route sanity test."""
    full_route_path= f'{TEST_CUSTOM_ROUTE_PREFIX}{TEST_CUSTOM_ROUTE_PATH}'
    async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
        response = await ac.get(full_route_path)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == full_route_path
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    expected_response_data = _get_response_data(CUSTOM_ROUTE_RETURN_VALUE)
    assert runner.resource['metadata']['Response Data'] == (
        expected_response_data
    )
    assert response_data == expected_response_data


@pytest.mark.asyncio
async def test_fastapi_exception(trace_transport, fastapi_app):
    """Test when the handler raises an exception."""
    try:
        async with AsyncClient(app=fastapi_app, base_url="http://test") as ac:
            _ = await ac.get("/err")
    except CustomFastAPIException:
        pass

    runner = trace_transport.last_trace.events[0]
    assert runner.error_code == ErrorCode.EXCEPTION
    assert runner.exception['type'] == 'CustomFastAPIException'
    assert runner.exception['message'] == 'test'
    assert runner.resource['metadata']['status_code'] == (
        DEFAULT_ERROR_STATUS_CODE
    )
    # validating no `zombie` traces exist
    assert not trace_factory.traces


@pytest.mark.asyncio
async def test_fastapi_handled_custom_exception(
        trace_transport,
        fastapi_app_with_exception_handlers
):
    """
    Test when the handler raises a custom exception and
    there's a matching exception handler.
    """
    app = fastapi_app_with_exception_handlers
    request_path = f'{HANDLED_EXCEPTION_PATH}?x=testval'
    async with AsyncClient(app=app, base_url="http://test") as ac:
        response = await ac.get(request_path)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == HANDLED_EXCEPTION_PATH
    assert runner.resource['metadata']['status_code'] == CUSTOM_STATUS_CODE
    assert runner.resource['metadata']['Response Data'] == (
        CUSTON_EXCEPTION_HANDLER_RESPONSE
    )
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert response_data == CUSTON_EXCEPTION_HANDLER_RESPONSE
    # validating no `zombie` traces exist
    assert not trace_factory.traces


@pytest.mark.asyncio
async def test_fastapi_handled_default_exception(
        trace_transport,
        fastapi_app_with_exception_handlers
):
    """
    Test when the handler raises Exception() and
    there's a matching default exception handler.
    """
    app = fastapi_app_with_exception_handlers
    request_path = f'{UNHANDLED_EXCEPTION_PATH}?x=testval'
    try:
        async with AsyncClient(app=app, base_url="http://test") as ac:
            response = await ac.get(request_path)
    except UnhandledFastAPIException:
        pass

    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.error_code == ErrorCode.EXCEPTION
    assert runner.exception['type'] == 'UnhandledFastAPIException'
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == UNHANDLED_EXCEPTION_PATH
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    assert runner.resource['metadata']['Response Data'] == (
        DEFAULT_EXCEPTION_HANDLER_RESPONSE
    )
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    # validating no `zombie` traces exist
    assert not trace_factory.traces



async def _send_request(app, path, trace_transport):
    """ Send request and validates its response & trace """
    request_path = f'/{path}'
    async with AsyncClient(app=app, base_url="http://test") as ac:
        response = await ac.get(request_path)
    response_data = response.json()
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == request_path
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    expected_response_data = _get_response_data(path)
    assert runner.resource['metadata']['Response Data'] == (
        expected_response_data
    )
    assert response_data == expected_response_data


@pytest.mark.asyncio
async def test_fastapi_multiple_requests(trace_transport, fastapi_app):
    """ Multiple requests test """
    for _ in range(3):
        await asyncio.gather(
            _send_request(fastapi_app, "a", trace_transport),
            _send_request(fastapi_app, "b", trace_transport)
        )
    # validating no `zombie` traces exist
    assert not trace_factory.traces


async def _send_async_request(app, path):
    async with AsyncClient(app=app, base_url="http://test") as ac:
        return await ac.get(path)


def test_fastapi_multiple_threads_route(trace_transport, fastapi_app):
    """
    Tests request to a route, which invokes multiple threads.
    Validating no `zombie` traces exist (fromn the callback invoked threads)
    """
    loop = None
    try:
        loop = asyncio.new_event_loop()
        response = loop.run_until_complete(
            _send_async_request(
                fastapi_app,
                f"{MULTIPLE_THREADS_ROUTE}?x=testval"
            )
        )
    finally:
        if loop:
            loop.close()
    response_data = response.json()
    # expects only 1 event. The new threads events shouldn't belong this trace
    assert len(trace_transport.last_trace.events) == 1
    runner = trace_transport.last_trace.events[0]
    assert isinstance(runner, FastapiRunner)
    assert runner.resource['name'].startswith('127.0.0.1')
    assert runner.resource['metadata']['Path'] == MULTIPLE_THREADS_ROUTE
    assert runner.resource['metadata']['status_code'] == DEFAULT_SUCCESS_STATUS_CODE
    expected_response_data = _get_response_data(MULTIPLE_THREADS_RETURN_VALUE)
    assert runner.resource['metadata']['Response Data'] == (
        expected_response_data
    )
    assert runner.resource['metadata']['Query Params'] == { 'x': 'testval'}
    assert response_data == expected_response_data
    # validating no `zombie` traces exist
    assert not trace_factory.traces



import mock
from epsagon.events.greengrasssdk import GreengrassEventFactory


@mock.patch('epsagon.trace.TraceFactory.add_event')
def test_sanity(add_event_mock):
    params = {
        'topic': 'name',
        'queueFullPolicy': True,
        'payload': 'test',
    }
    GreengrassEventFactory.create_event(None, None, None, params, None, None, None)
    add_event_mock.assert_called_once()
    event = add_event_mock.call_args_list[0].args[0]
    assert event.event_id.startswith('greengrass-')
    assert event.resource['name'] == 'name'
    assert event.resource['operation'] == 'publish'
    assert event.resource['metadata']['aws.greengrass.queueFullPolicy'] == True
    assert event.resource['metadata']['aws.greengrass.payload'] == 'test'



"""
Triggers for aws_lambda
"""
# pylint: disable=too-many-lines

from __future__ import absolute_import
from uuid import uuid4
from importlib import import_module
import hashlib
import json
from epsagon.utils import add_data_if_needed, parse_json, print_debug
from ..event import BaseEvent
from ..constants import EPSAGON_HEADER

# Conditionally importing boto3
TypeDeserializer = None  # pylint: disable=invalid-name
try:
    TypeDeserializer = (  # pylint: disable=invalid-name
        import_module('boto3.dynamodb.types').TypeDeserializer
    )
except ImportError:
    pass


class BaseLambdaTrigger(BaseEvent):
    """
    Represents base Lambda trigger
    """
    ORIGIN = 'trigger'


class JSONLambdaTrigger(BaseLambdaTrigger):
    """
    Represents JSON simple Lambda trigger
    """
    RESOURCE_TYPE = 'json'

    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """

        super(JSONLambdaTrigger, self).__init__(start_time)

        self.event_id = 'trigger-{}'.format(str(uuid4()))

        self.resource['name'] = 'trigger-{}'.format(context.function_name)
        self.resource['operation'] = self.RESOURCE_TYPE
        self.resource['metadata'] = {}

        add_data_if_needed(self.resource['metadata'], 'data', event)


class S3LambdaTrigger(BaseLambdaTrigger):
    """
    Represents S3 Lambda trigger
    """
    RESOURCE_TYPE = 's3'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """

        super(S3LambdaTrigger, self).__init__(start_time)

        self.event_id = 's3-trigger-{0}'.format(
            event['Records'][0]['responseElements']['x-amz-request-id']
        )
        self.resource['name'] = event['Records'][0]['s3']['bucket']['name']
        self.resource['operation'] = event['Records'][0]['eventName']

        self.resource['metadata'] = {
            'region': event['Records'][0]['awsRegion'],
            'request_parameters': event['Records'][0]['requestParameters'],
            'user_identity': event['Records'][0]['userIdentity'],
            'object_key': event['Records'][0]['s3']['object']['key'],
            'object_size': event['Records'][0]['s3']['object']['size'],
            'object_etag': event['Records'][0]['s3']['object'].get('eTag', ''),
            'object_sequencer': event['Records'][0]['s3']['object'][
                'sequencer'
            ],
            'x-amz-request-id': event['Records'][0]['responseElements'][
                'x-amz-request-id'
            ],
        }


class DynamoDBLambdaTrigger(BaseLambdaTrigger):
    """
    Represents DynamoDB Lambda trigger
    """
    RESOURCE_TYPE = 'dynamodb'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """

        super(DynamoDBLambdaTrigger, self).__init__(start_time)
        self.deserializer = TypeDeserializer() if TypeDeserializer else None

        record = event['Records'][0]
        self.event_id = record['eventID']
        self.resource['name'] = record['eventSourceARN'].split('/')[-3]
        self.resource['operation'] = record['eventName']

        if record['eventName'] == 'REMOVE':
            deserialized_item = self._deserialize_item(
                record['dynamodb']['Keys']
            )
        elif record['eventName'] == 'MODIFY':
            deserialized_item = self._deserialize_item(
                record['dynamodb']['Keys']
            )
        else:
            item = record['dynamodb']['NewImage']
            deserialized_item = self._deserialize_item(item)
            add_data_if_needed(self.resource['metadata'], 'New Image', item)

        self.resource['metadata'] = {
            'region': record['awsRegion'],
            'sequence_number': record['dynamodb']['SequenceNumber'],
        }

        if deserialized_item is not None:
            self.resource['metadata']['item_hash'] = hashlib.md5(
                json.dumps(deserialized_item, sort_keys=True).encode('utf-8')
            ).hexdigest()

    def _deserialize_item(self, item):
        """
        Deserialize DynamoDB Item in order to remove types definitions.
        :param item: The item to deserialize.
        :return: Deserialized item.
        """
        if self.deserializer is None:
            return None
        deserialized_item = item.copy()
        for key in item:
            try:
                deserialized_item[key] = self.deserializer.deserialize(
                    item[key]
                )
            except (TypeError, AttributeError):
                break
        return deserialized_item


class KinesisLambdaTrigger(BaseLambdaTrigger):
    """
    Represents Kinesis Lambda trigger
    """
    RESOURCE_TYPE = 'kinesis'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """

        super(KinesisLambdaTrigger, self).__init__(start_time)

        self.event_id = event['Records'][0]['eventID']
        self.resource['name'] = \
            event['Records'][0]['eventSourceARN'].split('/')[-1]
        self.resource['operation'] = \
            event['Records'][0]['eventName'].replace('aws:kinesis:', '')

        self.resource['metadata'] = {
            'region': event['Records'][0]['awsRegion'],
            'invoke_identity': event['Records'][0]['invokeIdentityArn'],
            'sequence_number': event['Records'][0]['kinesis']['sequenceNumber'],
            'partition_key': event['Records'][0]['kinesis']['partitionKey'],
            'record_count': len(event['Records'])
        }


class SNSLambdaTrigger(BaseLambdaTrigger):
    """
    Represents SNS Lambda trigger
    """
    RESOURCE_TYPE = 'sns'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """

        super(SNSLambdaTrigger, self).__init__(start_time)

        self.event_id = str(event['Records'][0]['Sns']['MessageId'])
        self.resource['name'] = \
            event['Records'][0]['EventSubscriptionArn'].split(':')[-2]
        self.resource['operation'] = str(event['Records'][0]['Sns']['Type'])

        self.resource['metadata'] = {
            'Notification Subject': str(event['Records'][0]['Sns']['Subject'])
        }
        message = str(event['Records'][0]['Sns']['Message'])
        add_data_if_needed(
            self.resource['metadata'],
            'Notification Message',
            message
        )
        print_debug('Initialized SNS Lambda trigger')


class SQSLambdaTrigger(BaseLambdaTrigger):
    """
    Represents SQS Lambda trigger.
    """
    RESOURCE_TYPE = 'sqs'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch).
        :param event: event dict from the entry point.
        :param context: the context dict from the entry point.
        """

        super(SQSLambdaTrigger, self).__init__(start_time)

        record = event['Records'][0]
        self.event_id = record['messageId']
        self.resource['name'] = record['eventSourceARN'].split(':')[-1]
        self.resource['operation'] = 'ReceiveMessage'

        self.resource['metadata'] = {
            'MD5 Of Message Body': record['md5OfBody'],
            'Sender ID': record['attributes']['SenderId'],
            'Approximate Receive Count': record['attributes'][
                'ApproximateReceiveCount'
            ],
            'Sent Timestamp': record['attributes']['SentTimestamp'],
            'Approximate First Receive Timestamp': record['attributes'][
                'ApproximateFirstReceiveTimestamp'
            ],
        }
        sqs_message_body = record['body']

        self.resource['metadata']['Number Of Messages'] = len(event['Records'])

        add_data_if_needed(
            self.resource['metadata'],
            'Message Body',
            str(sqs_message_body)
        )

        message_body = parse_json(sqs_message_body)
        if not message_body:
            return
        message_body_input = message_body.get('input')
        if not message_body_input:
            return
        if isinstance(message_body_input, dict):
            steps_dict = message_body_input.get('Epsagon')
            if not steps_dict:
                return
            self.resource['metadata'] = {'steps_dict': steps_dict}


class ProxyAPIGatewayLambdaTrigger(BaseLambdaTrigger):
    """
    Represents a Proxy API Gateway Lambda trigger.
    """
    RESOURCE_TYPE = 'api_gateway'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """

        super(ProxyAPIGatewayLambdaTrigger, self).__init__(start_time)
        default_request_context = {
            'requestId': str(uuid4()),
            'apiId': 'N/A',
            'stage': event.get('environment', 'N/A')
        }
        request_context = event.get('requestContext', default_request_context)
        self.event_id = request_context['requestId']
        self.resource['name'] = event['headers'].get(
            'Host',
            request_context.get('apiId')
        )

        self.resource['operation'] = event['httpMethod']

        query_params = event.get(
            'queryStringParameters',
            event.get('queryParams', 'N/A')
        )
        path_params = event.get(
            'pathParameters',
            event.get('pathParams', 'N/A')
        )

        self.resource['metadata'] = {
            'stage': request_context.get('stage', 'N/A'),
            'query_string_parameters': query_params,
            'path_parameters': path_params,
            'path': event.get('resource', 'N/A'),
        }

        add_data_if_needed(self.resource['metadata'], 'body', event['body'])
        add_data_if_needed(
            self.resource['metadata'],
            'headers',
            event['headers']
        )


class NoProxyAPIGatewayLambdaTrigger(BaseLambdaTrigger):
    """
    Represents an API Gateway Lambda trigger that uses the API Gateway method
    request passthrough template.
    """
    RESOURCE_TYPE = 'api_gateway_no_proxy'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """

        super(NoProxyAPIGatewayLambdaTrigger, self).__init__(start_time)

        self.event_id = event['context']['request-id']
        self.resource['name'] = event['params']['header'].get(
            'Host',
            event['context']['api-id']
        )
        self.resource['operation'] = event['context']['http-method']

        self.resource['metadata'] = {
            'stage': event['context']['stage'],
            'query_string_parameters': event['params']['querystring'],
            'path_parameters': event['params']['path'],
            'path': event['context']['resource-path'],
        }

        add_data_if_needed(
            self.resource['metadata'],
            'body',
            event['body-json']
        )
        add_data_if_needed(
            self.resource['metadata'],
            'headers',
            event['params']['header']
        )


class ElasticLoadBalancerLambdaTrigger(BaseLambdaTrigger):
    """
    Represents an Elastic Load Balancer trigger.
    """
    RESOURCE_TYPE = 'elastic_load_balancer'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """
        super(ElasticLoadBalancerLambdaTrigger, self).__init__(start_time)

        self.event_id = 'elb-{}'.format(str(uuid4()))
        self.resource['name'] = event['headers']['host']
        self.resource['operation'] = event['httpMethod']

        epsagon_trace_id = event['headers'].get(EPSAGON_HEADER)
        self.resource['metadata'] = {
            'http_trace_id': epsagon_trace_id,
            'query_string_parameters': event['queryStringParameters'],
            'target_group_arn': (
                event['requestContext']['elb']['targetGroupArn']
            ),
            'path': event['path']
        }

        if epsagon_trace_id:
            self.resource['metadata']['http_trace_id'] = epsagon_trace_id

        add_data_if_needed(
            self.resource['metadata'],
            'body',
            event['body']
        )
        add_data_if_needed(
            self.resource['metadata'],
            'headers',
            event['headers']
        )


class EventsLambdaTrigger(BaseLambdaTrigger):
    """
    Represents Events (schedule) Lambda trigger
    """
    RESOURCE_TYPE = 'events'

    # pylint: disable=W0613
    def __init__(self, start_time, event, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        """

        super(EventsLambdaTrigger, self).__init__(start_time)

        self.event_id = str(event['id'])
        name = 'CloudWatch Events'
        if event['resources'] and isinstance(event['resources'][0], str):
            name = str(event['resources'][0].split('/')[-1])
        self.resource['name'] = name
        self.resource['operation'] = str(event['detail-type'])

        self.resource['metadata'] = {
            'region': event['region'],
            'detail': None if not event['detail'] else event['detail'],
            'account': str(event['account']),
        }


class CognitoLambdaTrigger(BaseLambdaTrigger):
    """
    Represents a Cognito Lambda Trigger
    """
    RESOURCE_TYPE = 'cognito'

    def __init__(self, start_time, event, _context):
        """
        Initialize Trigger
        :param start_time: event's start time (epoch unix)
        :param event: event dict from entry point
        :param context: context dict from entry point
        """

        super(CognitoLambdaTrigger, self).__init__(start_time)

        self.event_id = 'cognito-{uid}'.format(uid=str(uuid4()))
        self.resource['name'] = event.get('userPoolId')
        self.resource['operation'] = event.get('triggerSource')

        self.resource['metadata'] = {
            'region': event.get('region'),
            'username': event.get('userName'),
            'client_id': event.get('callerContext', {}).get('clientId'),
            'user_pool_id': event.get('userPoolId'),
            'trigger_source': event.get('triggerSource'),
            'version': event.get('version')
        }

        add_data_if_needed(
            self.resource['metadata'],
            'attributes',
            event.get('request', {}).get('userAttributes')
        )

        add_data_if_needed(
            self.resource['metadata'],
            'session',
            event.get('request', {}).get('session')
        )

        add_data_if_needed(
            self.resource['metadata'],
            'callercontext',
            event.get('callerContext')
        )

        add_data_if_needed(
            self.resource['metadata'],
            'response',
            event.get('response'),
        )


class LambdaTriggerFactory(object):
    """
    Represents a Lambda Trigger Factory.
    """
    FACTORY = {
        class_obj.RESOURCE_TYPE: class_obj
        for class_obj in BaseLambdaTrigger.__subclasses__()
    }

    @staticmethod
    def factory(start_time, event, context):
        """
        Creates trigger event object.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param context: the context dict from the entry point
        :return: Event
        """
        trigger_service = JSONLambdaTrigger.RESOURCE_TYPE
        if 'Records' in event:
            event_source = 'eventSource'
            if event_source not in event['Records'][0]:
                event_source = 'EventSource'
            trigger_service = event['Records'][0][event_source].split(':')[-1]
        elif 'requestContext' in event and 'elb' in event['requestContext']:
            trigger_service = ElasticLoadBalancerLambdaTrigger.RESOURCE_TYPE
        elif 'httpMethod' in event:
            trigger_service = ProxyAPIGatewayLambdaTrigger.RESOURCE_TYPE
        elif 'context' in event and 'http-method' in event['context']:
            trigger_service = NoProxyAPIGatewayLambdaTrigger.RESOURCE_TYPE
        elif 'source' in event and 'detail-type' in event and 'detail' in event:
            trigger_service = EventsLambdaTrigger.RESOURCE_TYPE
        elif 'userPoolId' in event:
            trigger_service = CognitoLambdaTrigger.RESOURCE_TYPE
        elif 'source' in event:
            trigger_service = str(event['source'].split('.')[-1])
        return LambdaTriggerFactory.FACTORY.get(trigger_service)(
            start_time,
            event,
            context
        )



#!/usr/bin/env python
import re
import os
from setuptools import setup, find_packages

with open('./requirements.txt', 'r') as reqs_file:
    reqs = reqs_file.readlines()

# Get version
with open(os.path.join('epsagon', 'constants.py'), 'rt') as consts_file:
    version = re.search(r'__version__ = \'(.*?)\'', consts_file.read()).group(1)

setup(
    name='epsagon',
    version=version,
    description='Epsagon Instrumentation for Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Epsagon',
    author_email='support@epsagon.com',
    url='https://github.com/epsagon/epsagon-python',
    packages=find_packages(exclude=('tests', 'examples')),
    package_data={'epsagon': ['*.pem']},
    install_requires=reqs,
    license='MIT',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    entry_points={
        'epsagon': ['string = epsagon:auto_load']
    },
    keywords=[
        'serverless',
        'microservices',
        'epsagon',
        'tracing',
        'distributed-tracing',
        'lambda',
        'aws-lambda',
        'debugging',
        'monitoring'
    ],
    classifiers=(
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    )
)



import epsagon.wrappers.python_function
import epsagon.runners.python_function
import epsagon.constants
import mock
from tornado.httpclient import AsyncHTTPClient

TEST_URL = 'https://example.test/'


@mock.patch('epsagon.trace.TraceFactory.add_event')
def test_sanity(add_event_mock):
    retval = 'success'

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function():
        http_client = AsyncHTTPClient()
        http_client.fetch(TEST_URL)
        return retval
    assert wrapped_function() == retval
    add_event_mock.assert_called()
    event = add_event_mock.call_args_list[0].args[0]
    assert event.resource['name'] == 'example.test'
    assert event.resource['operation'] == 'GET'
    assert event.resource['type'] == 'http'
    assert 'http_trace_id' in event.resource['metadata']



import epsagon.wrappers.python_function
import epsagon.runners.python_function
import epsagon.constants
import mock
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


DB_NAME = 'db'
HOST_NAME = 'host'
ENGINE = create_engine('postgresql://user:password@{}/{}'.format(HOST_NAME, DB_NAME))


@mock.patch('epsagon.events.sqlalchemy.SqlAlchemyEvent')
def test_sanity(sqlalchemy_event_mock):
    retval = 'success'

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function():
        session = sessionmaker(bind=ENGINE)()
        session.close()
        return retval

    assert wrapped_function() == retval
    init_instrumentation, close_instrumentation = [
        call.args for call in sqlalchemy_event_mock.call_args_list
    ]

    assert init_instrumentation[-1] =='initialize'
    assert close_instrumentation[-1] == 'close'

    init_instrumentation[1].bind.url.database = DB_NAME
    init_instrumentation[1].bind.url.host = HOST_NAME
    close_instrumentation[1].bind.url.database = DB_NAME
    close_instrumentation[1].bind.url.host = HOST_NAME



"""
Runner for a Celery Python function
"""

from __future__ import absolute_import
import time
from uuid import uuid4
from importlib import import_module
from ..event import BaseEvent
from ..utils import add_data_if_needed


class CeleryRunner(BaseEvent):
    """
    Represents Python Celery event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'celery'
    OPERATION = 'execute'

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Initialize.
        :param start_time: event's start time (epoch).
        """

        super(CeleryRunner, self).__init__(time.time())

        self.event_id = str(uuid4())

        self.resource['name'] = (
            kwargs.get('sender').name
            if kwargs.get('sender')
            else ''
        )
        self.resource['operation'] = self.OPERATION

        app_conn = import_module('celery').current_app.connection()
        task_id = kwargs.get('task_id', '')
        body = kwargs.get('args')
        retval = kwargs.get('retval')
        state = kwargs.get('state', '')

        self.resource['metadata'].update({
            'id': task_id,
            'state': state,
            'hostname': app_conn.hostname or 'localhost',
            'virtual_host': app_conn.virtual_host,
            'driver': app_conn.transport.driver_type,
        })

        if body:
            add_data_if_needed(
                self.resource['metadata'],
                'args',
                body
            )

        if retval:
            add_data_if_needed(
                self.resource['metadata'],
                'retval',
                retval
            )

    def set_retry(self, attempt_number):
        """
        Setting retry attempt number
        """
        self.resource['metadata']['attempt_number'] = attempt_number



"""
kafka-python events.
"""

from __future__ import absolute_import
import traceback
from uuid import uuid4

from epsagon.utils import add_data_if_needed
from ..trace import trace_factory
from ..event import BaseEvent
from ..constants import EPSAGON_HEADER


class KafkaEvent(BaseEvent):
    """
    Represents base Kafka event.
    """

    ORIGIN = 'kafka'
    RESOURCE_TYPE = 'kafka'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(KafkaEvent, self).__init__(start_time)
        self.event_id = 'kafka-{}'.format(str(uuid4()))

        topic = args[0]
        headers = dict(kwargs['headers'])
        servers = instance.config['bootstrap_servers']
        if servers and isinstance(servers, list):
            # Take the first server if it is a list
            servers = servers[0]

        self.resource['name'] = topic
        self.resource['operation'] = 'send'
        self.resource['metadata'] = {
            'messaging.system': 'kafka',
            'messaging.destination': topic,
            'messaging.url': servers,
            'messaging.message_payload_size_bytes': (
                len(str(kwargs.get('value', '')))
            ),
        }
        if instance.config.get('client_id'):
            self.resource['metadata']['messaging.kafka.client_id'] = (
                instance.config['client_id']
            )
        if headers.get(EPSAGON_HEADER):
            self.resource['metadata'][EPSAGON_HEADER] = (
                headers[EPSAGON_HEADER]
            )
        if kwargs['key']:
            self.resource['metadata']['messaging.kafka.message_key'] = (
                kwargs['key']
            )

        add_data_if_needed(
            self.resource['metadata'],
            'messaging.headers',
            headers
        )

        add_data_if_needed(
            self.resource['metadata'],
            'messaging.message',
            kwargs['value']
        )

        if getattr(response, 'value', None) is not None:
            self.update_response(response.value)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """
        self.resource['metadata']['messaging.kafka.partition'] = (
            response.partition
        )


class KafkaEventFactory(object):
    """
    Factory class, generates a kafka event.
    """

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """Create an event"""
        event = KafkaEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        trace_factory.add_event(event)



"""
Wrapper for Google Function.
"""

from __future__ import absolute_import
import traceback
import time
import functools
import warnings
import epsagon.trace
import epsagon.wrappers.python_function
from epsagon.common import EpsagonWarning
from ..runners.gcp_function import GoogleFunctionRunner
from .. import constants


def gcp_wrapper(func):
    """Epsagon's GCP wrapper."""

    @functools.wraps(func)
    def _gcp_wrapper(*args, **kwargs):
        """
        Generic google function wrapper
        """
        trace = epsagon.trace.trace_factory.get_or_create_trace()
        trace.prepare()

        try:
            runner = GoogleFunctionRunner(
                time.time(),
            )
        # pylint: disable=W0703
        except Exception as exception:
            # Regress to python runner.
            warnings.warn(
                'GCP environment is invalid, using simple python wrapper',
                EpsagonWarning
            )
            trace.add_exception(exception, traceback.format_exc())
            return epsagon.wrappers.python_function.wrap_python_function(
                func,
                args,
                kwargs
            )

        constants.COLD_START = False

        result = None
        try:
            result = func(*args, **kwargs)
            return result
        # pylint: disable=W0703
        except Exception as exception:
            runner.set_exception(exception, traceback.format_exc())
            raise
        finally:
            try:
                if not trace.metadata_only:
                    runner.resource['metadata']['return_value'] = result
            # pylint: disable=W0703
            except Exception as exception:
                trace.add_exception(exception, traceback.format_exc())
            try:
                trace.add_event(runner)
                epsagon.trace.trace_factory.send_traces()
            # pylint: disable=W0703
            except Exception:
                pass

    return _gcp_wrapper



"""
Django patcher module.
"""

from __future__ import absolute_import
import wrapt
from ..utils import print_debug, is_lambda_env

try:
    from django.conf import settings
except ImportError:
    settings = None  # pylint: disable=invalid-name

EPSAGON_MIDDLEWARE = 'epsagon.wrappers.django.DjangoMiddleware'


def _wrapper(wrapped, _instance, args, kwargs):
    """
    Adds `EPSAGON_MIDDLEWARE` into django.conf.settings.
    :param wrapped: wrapt's wrapped
    :param _instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    """

    # Skip on Lambda environment since it's not relevant and might be duplicate
    if is_lambda_env():
        return wrapped(*args, **kwargs)

    try:
        # Extract middleware engine (varying between Django versions)
        if hasattr(settings, 'MIDDLEWARE') and settings.MIDDLEWARE is not None:
            # Check if not instrumented already
            if EPSAGON_MIDDLEWARE in settings.MIDDLEWARE:
                return wrapped(*args, **kwargs)

            # Add Epsagon's middleware to the list
            if isinstance(settings.MIDDLEWARE, tuple):
                settings.MIDDLEWARE = (
                    (EPSAGON_MIDDLEWARE,) + settings.MIDDLEWARE
                )
            elif isinstance(settings.MIDDLEWARE, list):
                settings.MIDDLEWARE = [EPSAGON_MIDDLEWARE] + settings.MIDDLEWARE
        elif (
            hasattr(settings, 'MIDDLEWARE_CLASSES') and
            settings.MIDDLEWARE_CLASSES is not None
        ):
            # Check if not instrumented already
            if EPSAGON_MIDDLEWARE in settings.MIDDLEWARE_CLASSES:
                return wrapped(*args, **kwargs)

            # Add Epsagon's middleware to the list
            if isinstance(settings.MIDDLEWARE_CLASSES, tuple):
                settings.MIDDLEWARE = (
                    (EPSAGON_MIDDLEWARE,) + settings.MIDDLEWARE_CLASSES
                )
            elif isinstance(settings.MIDDLEWARE_CLASSES, list):
                settings.MIDDLEWARE = (
                    [EPSAGON_MIDDLEWARE] + settings.MIDDLEWARE_CLASSES
                )
    except Exception:  # pylint: disable=broad-except
        print_debug('Could not add Django middleware')
    return wrapped(*args, **kwargs)


def patch():
    """
    Patch module.
    :return: None
    """
    wrapt.wrap_function_wrapper(
        'django.core.handlers.base',
        'BaseHandler.load_middleware',
        _wrapper
    )



"""
fastapi patcher module.
"""

from __future__ import absolute_import
import wrapt
from ..wrappers.fastapi import (
    exception_handler_wrapper,
    server_call_wrapper,
    route_class_wrapper,
)
from ..utils import is_lambda_env


def _exception_handler_wrapper(wrapped, _instance, args, kwargs):
    """
    Wraps the handler given to add_exception_handler.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    """
    # Skip on Lambda environment since it's not relevant and might be duplicate
    if is_lambda_env():
        return wrapped(*args, **kwargs)

    if args and len(args) == 2:
        args = list(args)
        args[1] = exception_handler_wrapper(args[1])
    return wrapped(*args, **kwargs)


def patch():
    """
    Patch module.
    :return: None
    """
    wrapt.wrap_function_wrapper(
        'fastapi.routing',
        'APIRoute.__init__',
        route_class_wrapper
    )
    wrapt.wrap_function_wrapper(
        'starlette.applications',
        'Starlette.add_exception_handler',
        _exception_handler_wrapper
    )
    wrapt.wrap_function_wrapper(
        'starlette.middleware.errors',
        'ServerErrorMiddleware.__call__',
        server_call_wrapper
    )



"""
Greengrass events module.
"""

from __future__ import absolute_import
import traceback
from uuid import uuid4

from epsagon.utils import add_data_if_needed
from ..trace import trace_factory
from ..event import BaseEvent


class GreengrassPublishEvent(BaseEvent):
    """
    Represents Greengrass publish event.
    """

    ORIGIN = 'greengrasssdk'
    RESOURCE_TYPE = 'greengrass'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(GreengrassPublishEvent, self).__init__(start_time)

        self.event_id = 'greengrass-{}'.format(str(uuid4()))

        self.resource['name'] = kwargs.get('topic', 'N/A')
        self.resource['operation'] = 'publish'
        if kwargs.get('queueFullPolicy'):
            self.resource['metadata']['aws.greengrass.queueFullPolicy'] = (
                kwargs.get('queueFullPolicy')
            )

        add_data_if_needed(
            self.resource['metadata'],
            'aws.greengrass.payload',
            kwargs.get('payload')
        )

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())


class GreengrassEventFactory(object):
    """
    Factory class, generates Greengrass event.
    """

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create an event according to the given api_name.
        """
        event = GreengrassPublishEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        trace_factory.add_event(event)



"""
Runner for a fastapi Python function
"""

from __future__ import absolute_import
import json
import uuid
import warnings
from fastapi.responses import (
    Response,
    JSONResponse,
    HTMLResponse,
    PlainTextResponse,
    UJSONResponse,
    RedirectResponse,
)
from fastapi.encoders import jsonable_encoder
from epsagon.common import EpsagonWarning
from ..event import BaseEvent
from ..utils import add_data_if_needed, normalize_http_url, print_debug
from ..constants import EPSAGON_HEADER

SUPPORTED_RAW_RESPONSE_TYPES = (
    JSONResponse,
    HTMLResponse,
    PlainTextResponse,
    UJSONResponse,
    RedirectResponse,
)

class FastapiRunner(BaseEvent):
    """
    Represents Python FastAPI event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'fastapi'
    OPERATION = 'request'

    def __init__(self, start_time, request):
        """
        Initialize.
        :param start_time: event's start time (epoch).
        :param request: the incoming request.
        """

        super(FastapiRunner, self).__init__(start_time)

        self.event_id = str(uuid.uuid4())

        self.resource['name'] = normalize_http_url(request.client.host)
        self.resource['operation'] = request.method

        self.resource['metadata'].update({
            'Base URL': str(request.base_url),
            'Path': request.url.path,
            'User Agent': request.headers.get('User-Agent', 'N/A'),
        })

        query_params = request.query_params
        if query_params:
            self.resource['metadata']['Query Params'] = dict(
                query_params.items()
            )

        request_headers = dict(request.headers.items())
        if request_headers.get(EPSAGON_HEADER):
            self.resource['metadata']['http_trace_id'] = request_headers.get(
                EPSAGON_HEADER
            )
        if request_headers:
            add_data_if_needed(
                self.resource['metadata'],
                'Request Headers',
                request_headers
            )

    def update_request_body(self, body):
        """
        Adds request body to event
        """
        if body:
            add_data_if_needed(
                self.resource['metadata'],
                'Request Data',
                body
            )

    def _update_raw_response_body(self, response, response_type):
        """
        Updates the response body by given `raw` response and its type
        """
        body = response.body
        if response_type == JSONResponse:
            try:
                body = json.loads(body)
            except Exception: # pylint: disable=W0703
                warnings.warn(
                    'Could not load response json',
                    EpsagonWarning
                )
                body = body.decode('utf-8')
        else:
            body = body.decode('utf-8')
        add_data_if_needed(
            self.resource['metadata'],
            'Response Data',
            body
        )


    def update_status_code(self, status_code, override=True):
        """
        Updates the event with given status code.
        :param override: indicates whether to override existing status code
        """
        if self.resource['metadata'].get('status_code') and not override:
            return
        self.resource['metadata']['status_code'] = status_code
        if status_code and status_code >= 500:
            self.set_error()

    def _update_raw_response(self, response):
        """
        Updates the event with data by given raw response.
        Raw response is an instance of Response.
        """
        for response_type in SUPPORTED_RAW_RESPONSE_TYPES:
            if isinstance(response, response_type):
                self._update_raw_response_body(response, response_type)
                break

        response_headers = dict(response.headers.items())
        if response.headers:
            add_data_if_needed(
                self.resource['metadata'],
                'Response Headers',
                response_headers
            )
        self.update_status_code(response.status_code)

    def update_response(self, response, status_code=None):
        """
        Adds response data to event.
        """
        if isinstance(response, Response):
            self._update_raw_response(response)
        else:
            try:
                add_data_if_needed(
                    self.resource['metadata'],
                    'Response Data',
                    jsonable_encoder(response)
                )
            except Exception: # pylint: disable=W0703
                print_debug(
                    'Could not json encode fastapi handler response data'
                )
            if status_code:
                self.update_status_code(status_code)
